from typing import Dict, Any, Callable, Type, TypeVar, Set

from pydantic import BaseModel, ValidationError

from .error import AutumnValidationError, AutumnHTTPError


T = TypeVar("T", bound=BaseModel)


def _decompose_value(value: Any) -> Any:
    if isinstance(value, BaseModel):
        return value.model_dump()
    elif isinstance(value, list):
        return [_decompose_value(item) for item in value]

    return value


def _build_payload(
    scope: Dict[str, Any], method: Callable, *, ignore: Set[str] = set()
) -> Dict[str, Any]:
    params = method.__code__.co_varnames
    payload: Dict[str, Any] = {}

    for key, value in scope.items():
        if key != "self" and key in params and key not in ignore and value is not None:
            payload[key] = _decompose_value(value)

    return payload


def _build_model(model: Type[T], data: Dict[str, Any]) -> T:
    try:
        return model.model_validate(data)
    except ValidationError as e:
        errors = e.errors()
        error_message = errors[0]["msg"]
        error_path = errors[0]["loc"]
        error_code = errors[0]["type"]

        err = AutumnValidationError(
            f"{error_message} at {error_path} with code {error_code}",
            "validation_error",
        )
        err.add_note("Received response: " + str(data))
        raise err


def _check_response(status_code: int, data: Dict[str, Any]) -> None:
    if not 200 <= status_code < 300:
        message = data.get("message", "No error message provided.")
        code = data.get("code", "unknown_error")
        raise AutumnHTTPError(
            message,
            code,
            status_code,
        )



from __future__ import annotations

from typing import (
    Optional,
    Dict,
    Any,
    TypeVar,
    Generic,
    Union,
    overload,
    Coroutine,
    TYPE_CHECKING,
)

from .types.customers import Customer
from .types.response import BillingPortalResponse, PricingTableResponse
from .utils import _build_payload

if TYPE_CHECKING:
    from .http import HTTPClient
    from .aio.http import AsyncHTTPClient

T_HttpClient = TypeVar("T_HttpClient", "AsyncHTTPClient", "HTTPClient")

__all__ = ("Customers",)


class Customers(Generic[T_HttpClient]):
    """An interface to Autumn's customer API.

    .. warning::
        This class is not intended for public use. It is used internally by the :class:`autumn.Client` class.
        Do not initialize this class directly.
    """

    def __init__(self, http: T_HttpClient):
        self._http = http

    @overload
    def get(self: "Customers[HTTPClient]", customer_id: str) -> Customer: ...

    @overload
    def get(
        self: "Customers[AsyncHTTPClient]", customer_id: str
    ) -> Coroutine[Any, Any, Customer]: ...

    def get(self, customer_id) -> Union[Customer, Coroutine[Any, Any, Customer]]:
        """Get a customer by their ID.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to get.

        Returns
        -------
        :class:`~autumn.types.customers.Customer`
            The customer.
        """
        return self._http.request("GET", f"/customers/{customer_id}", Customer)

    @overload
    def create(
        self: "Customers[HTTPClient]",
        id: str,
        *,
        email: Optional[str] = None,
        name: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Customer: ...

    @overload
    def create(
        self: "Customers[AsyncHTTPClient]",
        id: str,
        *,
        email: Optional[str] = None,
        name: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Coroutine[Any, Any, Customer]: ...

    def create(
        self,
        id: str,
        *,
        email: Optional[str] = None,
        name: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Union[Customer, Coroutine[Any, Any, Customer]]:
        """Create a new customer.

        |maybecoro|

        Parameters
        ----------
        id: str
            The ID of the customer to create.
        email: Optional[str]
            The customer's email address.
        name: Optional[str]
            The customer's name.
        metadata: Optional[Dict[str, Any]]
            Additional metadata to attach to the customer.

        Returns
        -------
        :class:`~autumn.types.customers.Customer`
            The created customer.
        """
        payload = _build_payload(locals(), self.create)  # type: ignore
        return self._http.request("POST", "/customers", Customer, json=payload)

    @overload
    def update(
        self: "Customers[HTTPClient]",
        customer_id: str,
        *,
        name: Optional[str] = None,
        email: Optional[str] = None,
        fingerprint: Optional[str] = None,
    ) -> Customer: ...

    @overload
    def update(
        self: "Customers[AsyncHTTPClient]",
        customer_id: str,
        *,
        name: Optional[str] = None,
        email: Optional[str] = None,
        fingerprint: Optional[str] = None,
    ) -> Coroutine[Any, Any, Customer]: ...

    def update(
        self,
        customer_id: str,
        *,
        name: Optional[str] = None,
        email: Optional[str] = None,
        fingerprint: Optional[str] = None,
    ) -> Union[Customer, Coroutine[Any, Any, Customer]]:
        """Update a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to update.
        name: Optional[str]
            The customer's new name.
        email: Optional[str]
            The customer's new email address.
        fingerprint: Optional[str]
            The customer's new fingerprint.

        Returns
        -------
        :class:`~autumn.types.customers.Customer`
            The updated customer.
        """
        payload = _build_payload(locals(), self.update, ignore={"customer_id"})  # type: ignore
        return self._http.request(
            "POST", f"/customers/{customer_id}", Customer, json=payload
        )

    @overload
    def get_billing_portal(
        self: "Customers[HTTPClient]",
        customer_id: str,
        *,
        return_url: Optional[str] = None,
    ) -> BillingPortalResponse: ...

    @overload
    def get_billing_portal(
        self: "Customers[AsyncHTTPClient]",
        customer_id: str,
        *,
        return_url: Optional[str] = None,
    ) -> Coroutine[Any, Any, BillingPortalResponse]: ...

    def get_billing_portal(
        self,
        customer_id: str,
        *,
        return_url: Optional[str] = None,
    ) -> Union[BillingPortalResponse, Coroutine[Any, Any, BillingPortalResponse]]:
        """Get a billing portal URL for a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to get a billing portal URL for.
        return_url: Optional[str]
            The URL to return to after the customer has completed the billing portal.

        Returns
        -------
        :class:`~autumn.types.response.BillingPortalResponse`
            The billing portal URL.
        """
        payload = _build_payload(
            locals(),
            self.get_billing_portal,  # type: ignore
            ignore={"customer_id"},
        )
        return self._http.request(
            "POST",
            f"/customers/{customer_id}/billing_portal",
            BillingPortalResponse,
            json=payload,
        )

    @overload
    def pricing_table(
        self: "Customers[HTTPClient]", customer_id: str
    ) -> PricingTableResponse: ...

    @overload
    def pricing_table(
        self: "Customers[AsyncHTTPClient]", customer_id: str
    ) -> Coroutine[Any, Any, PricingTableResponse]: ...

    def pricing_table(
        self, customer_id: str
    ) -> Union[PricingTableResponse, Coroutine[Any, Any, PricingTableResponse]]:
        """Get a pricing table for a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to get a pricing table for.

        Returns
        -------
        :class:`~autumn.types.response.PricingTableResponse`
            The pricing table.
        """
        params = {"customer_id": customer_id}
        return self._http.request(
            "GET",
            "/components/pricing_table",
            PricingTableResponse,
            params=params,
        )



from typing import Optional, List

from pydantic import BaseModel

from .meta import AppEnv
from .products import ProductItem, FreeTrial
from .customers import PricingTableProduct

__all__ = ("AttachResponse", "CheckResponse")


class AttachResponse(BaseModel):
    customer_id: str
    code: str = "No code returned."
    message: str
    checkout_url: Optional[str] = None
    product_ids: List[str]
    success: Optional[bool] = None


class CheckResponse(BaseModel):
    allowed: bool
    customer_id: str
    code: str
    balance: Optional[float] = None
    feature_id: Optional[str] = None
    product_id: Optional[str] = None
    status: Optional[str] = None
    # TODO: add preview attribute


class TrackResponse(BaseModel):
    id: str
    code: str
    customer_id: str
    feature_id: Optional[str] = None
    event_name: Optional[str] = None


class UsageResponse(BaseModel):
    code: str
    customer_id: str
    feature_id: str


class BillingPortalResponse(BaseModel):
    url: str
    customer_id: str


class CreateProductResponse(BaseModel):
    autumn_id: str
    created_at: int
    id: str
    name: str
    env: AppEnv
    is_add_on: bool
    is_default: bool
    items: List[ProductItem]
    free_trial: Optional[FreeTrial] = None


GetProductResponse = CreateProductResponse


class ReferralCodeResponse(BaseModel):
    code: str
    customer_id: str
    created_at: int


class ReferralRedeemResponse(BaseModel):
    id: str
    customer_id: str
    reward_id: str


class ProductCancelResponse(BaseModel):
    success: bool
    message: str
    customer_id: str
    product_id: str


class PricingTableResponse(BaseModel):
    list: List[PricingTableProduct]



from __future__ import annotations

from typing import TYPE_CHECKING, Tuple, Any

if TYPE_CHECKING:
    from starlette.requests import Request
    from ...aio.client import AsyncClient
    from ..app import AutumnIdentifyData


async def _extract(
    request: Request,
    *,
    get_json: bool = True,
) -> Tuple[AutumnIdentifyData, AsyncClient, Any]:
    autumn = getattr(request.state, "__autumn__")
    identify_func = autumn["identify"]
    client = autumn["client"]

    identify = await identify_func(request)
    json = await request.json() if get_json else None
    return (identify, client, json)



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from typing import Optional, List, Awaitable, Dict, Any

    from ..types.meta import AttachOption
    from ..types.features import Feature
    from ..types.meta import CustomerData
    from ..types.response import (
        AttachResponse,
        CheckResponse,
        TrackResponse,
    )


__all__ = ("AttachParams", "CheckParams", "TrackParams")


class AttachParams(Protocol):
    def __call__(
        self,
        customer_id: str,
        *,
        product_id: Optional[str] = None,
        product_ids: Optional[List[str]] = None,
        success_url: Optional[str] = None,
        force_checkout: bool = False,
        features: Optional[List[Feature]] = None,
        entity_id: Optional[str] = None,
        customer_data: Optional[CustomerData] = None,
        free_trial: Optional[bool] = None,
        options: Optional[List[AttachOption]] = None,
    ) -> Awaitable[AttachResponse]: ...


class CheckParams(Protocol):
    def __call__(
        self,
        customer_id: str,
        *,
        product_id: Optional[str] = None,
        feature_id: Optional[str] = None,
        required_balance: Optional[int] = 1,
        send_event: bool = False,
        with_preview: bool = False,
        entity_id: Optional[str] = None,
        customer_data: Optional[CustomerData] = None,
    ) -> Awaitable[CheckResponse]: ...


class TrackParams(Protocol):
    def __call__(
        self,
        customer_id: str,
        feature_id: str,
        *,
        value: int = 1,
        entity_id: Optional[str] = None,
        event_name: Optional[str] = None,
        idempotency_key: Optional[str] = None,
        properties: Optional[Dict[str, Any]] = None,
        customer_data: Optional[CustomerData] = None,
    ) -> Awaitable[TrackResponse]: ...



from enum import Enum
from typing import Literal, List, Optional, Union

from pydantic import BaseModel

__all__ = (
    "FreeTrialDuration",
    "UsageModel",
    "ProductItemInterval",
    "PriceCurrencyPair",
    "PriceTier",
    "ProductItem",
    "FreeTrial",
    "Product",
)


class FreeTrialDuration(str, Enum):
    DAY = "day"


class UsageModel(str, Enum):
    PREPAID = "prepaid"
    PAY_PER_USE = "pay_per_use"


class ProductItemInterval(str, Enum):
    MINUTE = "minute"
    HOUR = "hour"
    DAY = "day"
    WEEK = "week"
    MONTH = "month"
    QUARTER = "quarter"
    SEMI_ANNUAL = "semi_annual"
    YEAR = "year"
    MULTIPLE = "multiple"
    LIFETIME = "lifetime"


class PriceCurrencyPair(BaseModel):
    price: float
    currency: str


class PriceTier(BaseModel):
    to: float
    amount: Union[float, str]  # Can be a number or "inf"


class ProductItem(BaseModel):
    feature_id: Optional[str] = None
    included_usage: Optional[Union[float, str]] = None  # Infinite is "inf"
    interval: Optional[ProductItemInterval] = None
    usage_model: Optional[UsageModel] = None
    price: Optional[float] = None
    billing_units: Optional[float] = None
    entity_feature_id: Optional[str] = None
    reset_usage_on_billing: Optional[bool] = None
    reset_usage_when_enabled: Optional[bool] = None


class FreeTrial(BaseModel):
    duration: FreeTrialDuration
    length: float
    unique_fingerprint: bool


class Product(BaseModel):
    autumn_id: str
    created_at: float
    id: str
    name: str
    env: Literal["sandbox", "live"]
    is_add_on: bool
    is_default: bool
    group: str
    version: float
    items: List[ProductItem]
    free_trial: Optional[FreeTrial] = None



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from .http import AsyncHTTPClient
from ..client import Client
from ..features import Features
from ..customers import Customers
from ..products import Products

if TYPE_CHECKING:
    from typing_extensions import Self
    from .shed import AttachParams, CheckParams, TrackParams

__all__ = ("AsyncClient",)


class AsyncClient(Client):
    """
    The ``async`` client class for interacting with the Autumn API.

    This class is also exposed as ``autumn.Autumn``.

    Parameters
    ----------
    token: str
        The API key to use for authentication.
    base_url: Optional[str]
        The base URL of the Autumn API. This is useful when you are self-hosting Autumn and need to point to your own instance.

    Attributes
    ----------
    customers: :class:`~autumn.customers.Customers`
        An interface to Autumn's customer API.
    features: :class:`~autumn.features.Features`
        An interface to Autumn's feature API.
    products: :class:`~autumn.products.Products`
        An interface to Autumn's product API.

    """

    attach: AttachParams  # type: ignore
    check: CheckParams  # type: ignore
    track: TrackParams  # type: ignore

    def __init__(self, token: str, *, base_url: Optional[str] = None):
        from .. import BASE_URL, VERSION

        _base_url = base_url or BASE_URL
        _base_url = _base_url.rstrip("/")

        self.http = AsyncHTTPClient(_base_url, VERSION, token)
        self.customers = Customers(self.http)
        self.features = Features(self.http)
        self.products = Products(self.http)

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, _exc_type, _exc, _tb):
        await self.close()

    async def close(self):
        await self.http.close()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import sys
from pathlib import Path

sys.path.insert(0, str(Path(__file__).resolve().parents[2]))

project = "autumn-py"
copyright = "2025, Vish M"
author = "Vish M"
release = "1.3.2"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints",
    "sphinx_tabs.tabs",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]
html_theme_options = {}

rst_prolog: str = """
.. |maybecoro| replace:: This function returns an awaitable |coroutine_link|_ when using the :class:`~autumn.aio.client.AsyncClient` class.
.. |coroutine_link| replace:: *coroutine*
.. _coroutine_link: https://docs.python.org/3/library/asyncio-task.html#coroutine
"""

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "aiohttp": ("https://docs.aiohttp.org/en/stable", None),
    "pydantic": ("https://docs.pydantic.dev/latest", None),
    "starlette": ("https://www.starlette.io/", None),
}

resource_links = {
    "github": "https://github.com/justanotherbyte/autumn",
    "discord": "https://discord.gg/QDjfwGGWKT",
}


autodoc_typehints_description_target = "implementation"

pygments_style = "friendly"



from __future__ import annotations

from typing import (
    List,
    Any,
    Union,
    TypeVar,
    Generic,
    overload,
    Coroutine,
    Optional,
    TYPE_CHECKING,
)

from .types.products import ProductItem, FreeTrial
from .types.response import (
    CreateProductResponse,
    GetProductResponse,
    ReferralCodeResponse,
    ReferralRedeemResponse,
    ProductCancelResponse,
)
from .utils import _build_payload

if TYPE_CHECKING:
    from .http import HTTPClient
    from .aio.http import AsyncHTTPClient

T_HttpClient = TypeVar("T_HttpClient", "AsyncHTTPClient", "HTTPClient")


class Products(Generic[T_HttpClient]):
    """An interface to Autumn's product API.

    .. warning::
        This class is not intended for public use. It is used internally by the :class:`autumn.Client` class.
        Do not initialize this class directly.

    Example:

    .. code-block:: python

        import autumn

        client = autumn.Client(token="your_api_key")

        product = client.products.create(
            id="chat_messages",
            name="Chat Messages"
        )

        print(product.name) # Chat Messages

    """

    def __init__(self, http: T_HttpClient):
        self._http = http

    @overload
    def create(
        self: "Products[HTTPClient]",
        id: str,
        *,
        name: str,
        is_add_on: bool = False,
        is_default: bool = False,
        items: Optional[List[ProductItem]] = None,
        free_trial: Optional[FreeTrial] = None,
    ) -> CreateProductResponse: ...

    @overload
    def create(
        self: "Products[AsyncHTTPClient]",
        id: str,
        *,
        name: str,
        is_add_on: bool = False,
        is_default: bool = False,
        items: Optional[List[ProductItem]] = None,
        free_trial: Optional[FreeTrial] = None,
    ) -> Coroutine[Any, Any, CreateProductResponse]: ...

    def create(
        self,
        id: str,
        *,
        name: str,
        is_add_on: bool = False,
        is_default: bool = False,
        items: Optional[List[ProductItem]] = None,
        free_trial: Optional[FreeTrial] = None,
    ) -> Union[CreateProductResponse, Coroutine[Any, Any, CreateProductResponse]]:
        """Create a new product.

        |maybecoro|

        Parameters
        ----------
        id: str
            The ID of the product.
        name: str
            The name of the product.
        is_add_on: bool
            Whether the product is an add-on.
        is_default: bool
            Whether the product is a default product.
        items: Optional[List[ProductItem]]
            The items of the product.
        free_trial: Optional[FreeTrial]
            The free trial configuration of the product.

        Returns
        -------
        :class:`~autumn.types.response.CreateProductResponse`
            The response from the API.
        """
        payload = _build_payload(locals(), self.create)  # type: ignore
        return self._http.request(
            "POST", "/products", CreateProductResponse, json=payload
        )

    @overload
    def get(
        self: "Products[HTTPClient]",
        id: str,
    ) -> GetProductResponse: ...

    @overload
    def get(
        self: "Products[AsyncHTTPClient]",
        id: str,
    ) -> Coroutine[Any, Any, GetProductResponse]: ...

    def get(
        self, id: str
    ) -> Union[GetProductResponse, Coroutine[Any, Any, GetProductResponse]]:
        """Get a product by its ID.

        |maybecoro|

        Parameters
        ----------
        id: str
            The ID of the product to get.

        Returns
        -------
        :class:`~autumn.types.response.GetProductResponse`
            The response from the API.
        """
        return self._http.request("GET", f"/products/{id}", GetProductResponse)

    @overload
    def get_referral_code(
        self: "Products[HTTPClient]",
        customer_id: str,
        program_id: str,
    ) -> ReferralCodeResponse: ...

    @overload
    def get_referral_code(
        self: "Products[AsyncHTTPClient]",
        customer_id: str,
        program_id: str,
    ) -> Coroutine[Any, Any, ReferralCodeResponse]: ...

    def get_referral_code(
        self,
        customer_id: str,
        program_id: str,
    ) -> Union[ReferralCodeResponse, Coroutine[Any, Any, ReferralCodeResponse]]:
        """Get a referral code for a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to get a referral code for.
        program_id: str
            The ID of the program to get a referral code for.

        Returns
        -------
        :class:`~autumn.types.response.ReferralCodeResponse`
            The response from the API.
        """

        payload = _build_payload(locals(), self.get_referral_code)  # type: ignore
        return self._http.request(
            "POST",
            "/referrals/code",
            ReferralCodeResponse,
            json=payload,
        )

    @overload
    def redeem_referral_code(
        self: "Products[HTTPClient]",
        code: str,
        customer_id: str,
        reward_id: str,
    ) -> ReferralRedeemResponse: ...

    @overload
    def redeem_referral_code(
        self: "Products[AsyncHTTPClient]",
        code: str,
        customer_id: str,
        reward_id: str,
    ) -> Coroutine[Any, Any, ReferralRedeemResponse]: ...

    def redeem_referral_code(
        self,
        code: str,
        customer_id: str,
        reward_id: str,
    ) -> Union[ReferralRedeemResponse, Coroutine[Any, Any, ReferralRedeemResponse]]:
        """Redeem a referral code for a customer.

        |maybecoro|

        Parameters
        ----------
        code: str
            The code to redeem.
        customer_id: str
            The ID of the customer to redeem the code for.
        reward_id: str
            The ID of the reward to redeem.

        Returns
        -------
        :class:`~autumn.types.response.ReferralRedeemResponse`
            The response from the API.
        """
        payload = _build_payload(locals(), self.redeem_referral_code)  # type: ignore
        return self._http.request(
            "POST",
            "/referrals/redeem",
            ReferralRedeemResponse,
            json=payload,
        )

    @overload
    def cancel(
        self: "Products[HTTPClient]",
        customer_id: str,
        product_id: str,
        *,
        entity_id: Optional[str] = None,
        cancel_immediately: bool = False
    ) -> ProductCancelResponse: ...

    @overload
    def cancel(
        self: "Products[AsyncHTTPClient]",
        customer_id: str,
        product_id: str,
        *,
        entity_id: Optional[str] = None,
        cancel_immediately: bool = False
    ) -> Coroutine[Any, Any, ProductCancelResponse]: ...

    def cancel(
        self,
        customer_id: str,
        product_id: str,
        *,
        entity_id: Optional[str] = None,
        cancel_immediately: bool = False
    ) -> Union[ProductCancelResponse, Coroutine[Any, Any, ProductCancelResponse]]:
        """Cancel a product for a customer.

        |maybecoro|

        Parameters
        ----------
        customer_id: str
            The ID of the customer to cancel the product for.
        product_id: str
            The ID of the product to cancel.
        entity_id: Optional[str]
            The ID of the entity to cancel the product for.
        cancel_immediately: bool
            Whether to cancel the product immediately. If false, the product will be cancelled at the end of the billing cycle.

        Returns
        -------
        :class:`~autumn.types.response.ProductCancelResponse`
            The response from the API.
        """
        payload = _build_payload(locals(), self.cancel_product)  # type: ignore
        return self._http.request(
            "POST",
            "/cancel",
            ProductCancelResponse,
            json=payload,
        )



from .client import *
from .error import *
from .types.meta import *
from .types.balance import *
from .types.customers import *
from .types.features import *
from .types.products import *
from .aio.client import AsyncClient as Autumn


__version__ = "1.3.2.post1"
__license__ = "MIT"
__author__ = "justanotherbyte"
__copyright__ = "Copyright 2025 justanotherbyte"


BASE_URL = "https://api.useautumn.com"
VERSION = "v1"



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Optional

from pydantic import BaseModel

if TYPE_CHECKING:
    from typing import Dict, Any

    Metadata = Dict[str, Any]


__all__ = ("Empty", "AppEnv", "AttachOption", "CustomerData")


class Empty(BaseModel): ...


class AppEnv(str, Enum):
    PRODUCTION = "production"
    SANDBOX = "sandbox"


class AttachOption(BaseModel):
    feature_id: str
    quantity: int


class CustomerData(BaseModel):
    name: Optional[str] = None
    email: Optional[str] = None
    fingerprint: Optional[str] = None



from __future__ import annotations

from typing import TYPE_CHECKING

from starlette.responses import JSONResponse

from . import _extract
from ...types.meta import CustomerData

if TYPE_CHECKING:
    from starlette.requests import Request


async def attach_route(request: Request):
    identify, autumn, json = await _extract(request)

    customer_id = identify["customer_id"]
    customer_data = identify["customer_data"]

    response = await autumn.attach(
        customer_id, customer_data=CustomerData(**customer_data), **json
    )
    return JSONResponse(response.model_dump(mode="json"))


async def check_route(request: Request):
    identify, autumn, json = await _extract(request)

    customer_id = identify["customer_id"]
    response = await autumn.check(customer_id, **json)
    return JSONResponse(response.model_dump(mode="json"))


async def track_route(request: Request):
    identify, autumn, json = await _extract(request)

    customer_id = identify["customer_id"]
    response = await autumn.track(customer_id, **json)
    return JSONResponse(response.model_dump(mode="json"))


async def cancel_route(request: Request):
    identify, autumn, json = await _extract(request)

    customer_id = identify["customer_id"]
    response = await autumn.products.cancel(customer_id, **json)
    return JSONResponse(response.model_dump(mode="json"))


async def billing_portal_route(request: Request):
    identify, autumn, json = await _extract(request)

    customer_id = identify["customer_id"]
    response = await autumn.customers.get_billing_portal(customer_id, **json)
    return JSONResponse(response.model_dump(mode="json"))



from typing import Type, TypeVar

from pydantic import BaseModel

from ..error import AutumnError
from ..http import HTTPClient
from ..utils import _build_model, _check_response


try:
    import aiohttp
except ImportError:
    raise AutumnError(
        "aiohttp is not installed. Please install it with `pip install aiohttp`",
        "missing_dependency",
    )

T = TypeVar("T", bound=BaseModel)


__all__ = ("AsyncHTTPClient",)


class AsyncHTTPClient:
    def __init__(self, base_url: str, version: str, token: str):
        self.base_url = base_url
        self.version = version
        self.session = None  # type: ignore
        self._headers = HTTPClient._build_headers(token)

        self._build_url = HTTPClient._build_url

    def _ensure_session(self):
        if self.session is None:
            self.session = aiohttp.ClientSession()

    async def request(self, method: str, path: str, type_: Type[T], **kwargs) -> T:
        self._ensure_session()
        assert self.session is not None  # appease type checker

        url = self._build_url(self.base_url, self.version, path)

        async with self.session.request(
            method, url, headers=self._headers, **kwargs
        ) as resp:
            data = await resp.json()

        _check_response(resp.status, data)
        return _build_model(type_, data)

    async def close(self):
        if self.session is not None:
            await self.session.close()
